/**
* @file RectangleOperationPool.cpp
* @brief pool for rectangle feature extraction methods
* @author Sven Sickert
* @date 10/09/2015

*/

#include "RectangleOperationPool.h"

using namespace OBJREC;

RectangleOperationPool::RectangleOperationPool ( const NICE::Config *_config )
    : OperationPool ( _config )
{
    contextMode = false;
}

RectangleOperationPool::RectangleOperationPool ( const NICE::Config *_config,
                                           const bool _contextMode )
    : OperationPool ( _config )
{
    contextMode = _contextMode;
}

RectangleOperationPool::~RectangleOperationPool ( )
{

}

void RectangleOperationPool::getOperations ( std::string confString )
{
    if ( config->gB ( confString, "int", true ) )
    {
        Operation3D* o = new IntegralOps3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "bi_int", true ) )
    {
        Operation3D* o = new BiIntegralOps3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "bi_int_cent", true ) )
    {
        Operation3D* o = new BiIntegralCenteredOps3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "int_cent", true ) )
    {
        Operation3D* o = new IntegralCenteredOps3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_horz", true ) )
    {
        Operation3D* o = new HaarHorizontal3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_vert", true ) )
    {
        Operation3D* o = new HaarVertical3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_stack", true ) )
    {
        Operation3D* o = new HaarStacked3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_diagxy", true ) )
    {
        Operation3D* o = new HaarDiagXY3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_diagxz", true ) )
    {
        Operation3D* o = new HaarDiagXZ3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar_diagyz", true ) )
    {
        Operation3D* o = new HaarDiagYZ3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar3_horz", true ) )
    {
        Operation3D* o = new Haar3Horiz3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar3_vert", true ) )
    {
        Operation3D* o = new Haar3Vert3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "haar3_stack", true ) )
    {
        Operation3D* o = new Haar3Stack3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "local_mean", false ) )
    {
        Operation3D* o = new LocalMean3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "local_stat_mom", false ) )
    {
        Operation3D* o = new LocalStatMom3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
}
