/**
* @file SimpleOperationPool.cpp
* @brief pool for simple feature extraction methods
* @author Sven Sickert
* @date 10/09/2015

*/

#include "SimpleOperationPool.h"

using namespace OBJREC;

SimpleOperationPool::SimpleOperationPool ( )
    : OperationPool ( )
{
}

SimpleOperationPool::SimpleOperationPool ( const NICE::Config *_config )
    : OperationPool ( _config )
{
    contextMode = false;
}

SimpleOperationPool::SimpleOperationPool ( const NICE::Config *_config,
                                           const bool _contextMode )
    : OperationPool ( _config )
{
    contextMode = _contextMode;
}

SimpleOperationPool::~SimpleOperationPool ( )
{

}

void SimpleOperationPool::getOperations ( std::string confString )
{
    if ( config->gB ( confString, "minus", true ) )
    {
        Operation3D* o = new Minus3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "minus_abs", true ) )
    {
        Operation3D* o = new MinusAbs3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "addition", true ) )
    {
        Operation3D* o = new Addition3D();
        o->setContext( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "only1", true ) )
    {
        Operation3D* o = new Only13D();
        o->setContext ( contextMode );
        pool.push_back ( o );
    }
    if ( config->gB ( confString, "rel_x", true ) )
        pool.push_back ( new RelativeXPosition3D() );
    if ( config->gB ( confString, "rel_y", true ) )
        pool.push_back ( new RelativeYPosition3D() );
    if ( config->gB ( confString, "rel_z", true ) )
        pool.push_back ( new RelativeZPosition3D() );
}
